import json
from io import StringIO
from yaml import safe_dump, safe_load

from typing import List, Dict, Callable, Type, Any, TextIO, Optional, Literal

from .serialize import serialize_namedtuple, deserialize_namedtuple, PrimitiveType
from .typehelpers import NT, T


Format = Literal["json", "yaml"]


def _pretty_print(kwargs: Dict[str, Any]) -> Dict[str, Any]:
    if "indent" not in kwargs:
        kwargs["indent"] = "    "
    return kwargs


def namedtuple_sequence_dumps(
    nt_items: List[NT],
    *,
    attr_serializers: Optional[Dict[str, Callable[[T], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    format: Optional[Format] = "json",
    **kwargs: Any,
) -> str:
    """
    Dump the list of namedtuples to a JSON string
    """
    s_obj: List[Dict[str, Any]] = []
    for nt in nt_items:
        s_obj.append(serialize_namedtuple(nt, attr_serializers, type_serializers))
    if format == "json":
        return json.dumps(s_obj, **_pretty_print(kwargs))
    elif format == "yaml":
        buf = StringIO()
        safe_dump(s_obj, buf, **kwargs)
        return buf.getvalue()
    else:
        raise ValueError(
            f"Format is None while trying to dump {nt_items[0] if len(nt_items) > 0 else nt_items}"
        )


def namedtuple_sequence_dump(
    nt_items: List[NT],
    fp: TextIO,
    *,
    attr_serializers: Optional[Dict[str, Callable[[T], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    format: Optional[Format] = "json",
    **kwargs: Any,
) -> None:
    """
    Dump the list of namedtuples to a file-like object as JSON
    """
    # dump to string first, so JSON serialization errors dont cause data losses
    dumped: str = namedtuple_sequence_dumps(
        nt_items,
        attr_serializers=attr_serializers,
        type_serializers=type_serializers,
        format=format,
        **kwargs,
    )
    fp.write(dumped)


def _load_json(nt_string: str) -> Any:
    try:
        # speedup load if orjson is installed
        import orjson

        return orjson.loads(nt_string)
    except ImportError:
        pass
    return json.loads(nt_string)


def namedtuple_sequence_loads(
    nt_string: str,
    to: Type[NT],
    *,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    format: Optional[Format] = "json",
) -> List[NT]:
    """
    Load a list of namedtuples specificed by 'to' from a JSON string
    """
    if format == "json":
        loaded_obj = _load_json(nt_string)
    elif format == "yaml":
        loaded_obj = safe_load(nt_string)
    else:
        raise ValueError("unset format while trying to dump")
    if not isinstance(loaded_obj, list):
        raise TypeError(
            f"{loaded_obj} is a {type(loaded_obj).__name__}, expected a top-level list from JSON source"
        )
    ds_items: List[NT] = []
    for lo in loaded_obj:
        ds_items.append(
            deserialize_namedtuple(
                lo,
                to,
                attr_deserializers=attr_deserializers,
                type_deserializers=type_deserializers,
            )
        )
    return ds_items


def namedtuple_sequence_load(
    fp: TextIO,
    to: Type[NT],
    *,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    format: Optional[Format] = "json",
) -> List[NT]:
    """
    Load a list of namedtuples to the namedtuple specificed by 'to'
    from a file-like object containing JSON
    """
    return namedtuple_sequence_loads(
        fp.read(),
        to,
        attr_deserializers=attr_deserializers,
        type_deserializers=type_deserializers,
        format=format,
    )



from typing import NamedTuple
from datetime import timedelta
from autotui import (
    prompt_namedtuple,
    namedtuple_sequence_dumps,
    namedtuple_sequence_loads,
    AutoHandler,
)

# handle validating the user input
# can throw a ValueError
def _timedelta(user_input: str) -> timedelta:
    if len(user_input.strip()) == 0:
        raise ValueError("Not enough input!")
    minutes, _, seconds = user_input.partition(":")
    # could throw ValueError
    return timedelta(minutes=float(minutes), seconds=float(seconds))


# serializer for timedelta, converts to JSON-compatible integer
def to_seconds(t: timedelta) -> int:
    return int(t.total_seconds())


# deserializer from integer to timedelta
def from_seconds(seconds: int) -> timedelta:
    return timedelta(seconds=seconds)


# The data we want to persist to the file
class Action(NamedTuple):
    name: str
    duration: timedelta


# AutoHandler describes what function to use to validate
# user input, and which errors to wrap while validating
timedelta_handler = AutoHandler(
    func=_timedelta,  # accepts the string the user is typing as input
    catch_errors=[ValueError],
)

# Note: validators are of type
# Dict[Type, AutoHandler]
# serializer/deserializers are
# Dict[Type, Callable]
# the Callable accepts one argument,
# which is either the python value being serialized
# or the JSON value being deserialized

# use the validator to prompt the user for the NamedTuple data
# name: str automatically uses a generic string prompt
# duration: timedelta gets handled by the type_validator
a = prompt_namedtuple(
    Action,
    type_validators={
        timedelta: timedelta_handler,
    },
)


# Note: this specifies timedelta as the type,
# not int. It uses what the NamedTuple
# specifies as the type for that field, not
# the type of the value thats loaded from JSON

# dump to JSON
a_str: str = namedtuple_sequence_dumps(
    [a],
    type_serializers={
        timedelta: to_seconds,
    },
    indent=None,
)

# load from JSON
a_load = namedtuple_sequence_loads(
    a_str,
    to=Action,
    type_deserializers={
        timedelta: from_seconds,
    },
)[0]

# can also specify with attributes instead of types
a_load2 = namedtuple_sequence_loads(
    a_str,
    to=Action,
    attr_deserializers={
        "duration": from_seconds,
    },
)[0]

print(a)
print(a_str)
print(a_load)
print(a_load2)



import sys
import typing
import inspect
from functools import lru_cache
import types
from typing import (
    Tuple,
    cast,
    Optional,
    TypeVar,
    Callable,
    Type,
    Union,
    List,
    Any,
    Set,
    Sequence,
    Dict,
)
from enum import Enum

# namedtuple type - can't really bind this to anything, since tuple
# is too generic, and NamedTuple isn't a type
NT = TypeVar("NT")

# Generic type -- used elsewhere for function bindings
T = TypeVar("T")

# something with no arguments which when called
# prompts the user -- returns some value
PromptFunction = Callable[[], T]
OptionalPromptFunction = Callable[[], Optional[T]]
PromptFunctionorValue = Union[PromptFunction, T]

# A lot of these are helpers from:
# https://github.com/karlicoss/cachew/blob/f4db4a6c6609170642c6cd09d50b52ac4c0edec9/src/cachew/__init__.py#L144

# items that can serialized directly into JSON by json.dumps
PRIMITIVES = {
    str: Type[str],
    int: Type[int],
    float: Type[float],
    bool: Type[bool],
}

PrimitiveType = Union[str, int, float, bool, dict]


CONTAINERS = {
    list: Type[List],
    set: Type[Set],
}


def cache(user_function):
    'Simple lightweight unbounded cache. Sometimes called "memoize".'
    return lru_cache(maxsize=None)(user_function)


# needed to check if we can type hint generics
# https://www.python.org/dev/peps/pep-0585/
# or have unions like X | Y
# https://www.python.org/dev/peps/pep-0604/
above_39 = sys.version_info.major >= 3 and sys.version_info.minor >= 9
above_310 = sys.version_info.major >= 3 and sys.version_info.minor >= 10

AllowedContainers = Union[List[T], Set[T]]


def add_to_container(container: AllowedContainers, item: T) -> AllowedContainers:
    if isinstance(container, list):
        container.append(item)
    elif isinstance(container, set):
        container.add(item)
    else:
        raise RuntimeError(
            f"{type(container)} is not a list/set, not sure how to add to"
        )
    return container


@cache
def get_union_args(cls: Type) -> Optional[Tuple[List[Type[Any]], bool]]:
    """
    >>> get_union_args(Union[str, int])
    ([<class 'str'>, <class 'int'>], False)
    >>> get_union_args(Optional[str])
    ([<class 'str'>], True)
    >>> get_union_args(str)
    """
    is_union_type = False
    if above_310 and typing.get_origin(cls) == types.UnionType:  # type: ignore[attr-defined]
        is_union_type = True
    if not is_union_type:
        origin_attr = getattr(cls, "__origin__", None)
        if origin_attr == Union:
            is_union_type = True

    if not is_union_type:
        return None

    args: Type = cls.__args__
    arg_list: List[Type] = [e for e in args if e != type(None)]
    is_opt = type(None) in args
    assert len(arg_list) > 0
    return arg_list, is_opt


def resolve_annotation_single(cls: Type) -> Tuple[Type, bool]:
    """
    Given the annotation type from a namedtuple, extract the type
    Doesnt allow Unions other than Optional/None Unions
    """
    is_optional = False
    res = get_union_args(cls)
    if res is not None:
        attr_types, is_optional = res
        if len(attr_types) == 1:
            cls = attr_types[0]
    return cls, is_optional


# an estimation, someone could always trick
# this if they really wanted to...
# this should be passed the NamedTuple object,
# not a class type
def is_namedtuple_obj(thing: Any) -> bool:
    _asdict = getattr(thing, "_asdict", None)
    return _asdict is not None and callable(_asdict)


# this should be passed a NamedTuple type (typically
# created with class(NamedTuple), not an instance
def is_namedtuple_type(thing: Type) -> bool:
    return hasattr(thing, "_fields") and issubclass(thing, tuple) and callable(thing)


@cache
def enum_getval(enum: Type[Enum], value: Any) -> Enum:
    """
    Given some value and an enum, get the corresponding Enum value

    Prefer the enum value to the enum attribute name
    since its more likely the value is dumped to JSON
    """
    try:
        return enum[value]
    except KeyError:
        pass
    try:
        return enum(value)
    except ValueError:
        pass
    raise ValueError(
        f"Could not find {value} on Enumeration {enum} {enum.__members__.items()}"
    )


def is_union(cls: Type) -> bool:
    return get_union_args(cls) is not None


def is_optional(cls: Type) -> bool:
    res = get_union_args(cls)
    if res is None:
        return False
    _, is_opt = res
    assert isinstance(is_opt, bool)
    return is_opt


@cache
def get_collection_types(cls: Type) -> Tuple[Type, Type]:
    """
    >>> from typing import List
    >>> get_collection_types(List[int])
    (<class 'list'>, <class 'int'>)
    >>> get_collection_types(Set[bool])
    (<class 'set'>, <class 'bool'>)
    """
    container_type: Type = strip_generic(cls)
    # e.g. if List[int], internal[0] == int
    internal: Sequence[Type] = typing.get_args(cls)  # requires 3.8
    assert (
        len(internal) == 1
    ), f"Expected 1 argument for {container_type}, got {len(internal)}"
    return container_type, internal[0]


@cache
def strip_generic(tp):
    """
    >>> from typing import List
    >>> strip_generic(List[int])
    <class 'list'>
    >>> strip_generic(str)
    <class 'str'>
    """
    GA = getattr(typing, "_GenericAlias")
    if isinstance(tp, GA):
        return tp.__origin__
    if above_39:  # >= Python3.9
        origin = typing.get_origin(tp)
        if origin is not None:
            return origin
    return tp


def is_primitive(cls: Type) -> bool:
    """
    Whether or not this is a supported, serializable primitive

    >>> from typing import Dict
    >>> is_primitive(int)
    True
    >>> is_primitive(str)
    True
    >>> is_primitive(float)
    True
    >>> is_primitive(bool)
    True
    >>> is_primitive(set)
    False
    """
    return cls in PRIMITIVES


def is_supported_container(cls: Type) -> bool:
    """
    >>> from typing import Dict, List, Set, Tuple
    >>> is_supported_container(List[int])
    True
    >>> is_supported_container(Set[str])
    True
    >>> is_supported_container(Tuple[int])
    False
    >>> is_supported_container(Dict[int, str])
    False
    """
    return strip_generic(cls) in CONTAINERS


@cache
def inspect_signature_dict(nt: Callable[..., Any]) -> Dict[str, Type]:
    return {
        name: param.annotation
        for name, param in inspect.signature(nt).parameters.items()
    }


@cache
def enum_attribute_dict(enum_cls: Type[Enum]) -> Dict[str, Enum]:
    if not hasattr(enum_cls, "__members__"):
        raise TypeError(
            f"Could not find __members__ attribute on Enumeration {enum_cls}. May have passed a value instead of a type?"
        )
    return dict(enum_cls.__members__.items())



from typing import Dict, Type, Callable, Any, Union, Optional
from datetime import datetime, timezone
from enum import Enum

from .typehelpers import (
    is_supported_container,
    get_collection_types,
    is_primitive,
    resolve_annotation_single,
    PrimitiveType,
    inspect_signature_dict,
    is_namedtuple_type,
    enum_getval,
    NT,
    T,
)
from .warn import warn


def _serialize_type(
    value: Any,
    cls: Type,
    is_optional: bool,
    type_serializers: Dict[Type, Callable[[Any], PrimitiveType]],
) -> Optional[Union[PrimitiveType, Any]]:
    """
    Gets one of the built-in serializers or a type_serializers from the user,
    and serializes the value from the NamedTuple to that
    """
    # use type serializers first, user may have already specified some way
    # to handle nulls using a custom function
    if cls in type_serializers:
        return type_serializers[cls](value)
    # value can still be None here, we checked against namedtuple field type, not the dynamic
    # type of the value given
    if value is None:
        if not is_optional:
            warn(
                f"No value for non-optional type {value}, attempting to be serialized to {cls.__name__}"
            )
        return None  # serialized to null
    else:
        if cls == datetime:
            return int(value.timestamp())
        elif issubclass(cls, Enum):
            # assumes that the enumeration value the user provided is JSON-serializable
            if isinstance(value, Enum):
                # https://docs.python.org/3/library/enum.html#programmatic-access-to-enumeration-members-and-their-attributes
                return value.value
            else:
                # this isn't an enum type, but the user specified the direct value, so we're
                # assuming that its a valid value for this enumeration. The value is dumped,
                # and next time its loaded it should deserialize onto the Enum value
                return value
        elif is_primitive(cls):
            return value  # all other primitives are JSON compatible
        elif is_namedtuple_type(cls):
            # if the attribute for this value is another NamedTuple,
            # recursively serialize the value
            return serialize_namedtuple(value, type_serializers=type_serializers)
    warn(f"No known way to serialize {cls.__name__}")
    return value
    # raise? it'll fail when json module fails to do it anyways, so
    # might as well leave it
    # raise AutoTUIException(f"no known way to serialize {cls}")


def serialize_namedtuple(
    nt: NT,
    attr_serializers: Optional[Dict[str, Callable[[T], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[T], PrimitiveType]]] = None,
) -> Dict[str, Any]:
    """
    Serializes a NamedTuples to a JSON-compatible dictionary

    If the user provides attr_serializers or type_serializers, uses those
    instead of the defaults.
    """
    attr_serializers = attr_serializers or {}
    type_serializers = type_serializers or {}

    json_dict: Dict[str, Any] = {}

    for attr_name, nt_annotation in inspect_signature_dict(nt.__class__).items():

        # (<class 'int'>, False)
        attr_type, is_optional = resolve_annotation_single(nt_annotation)

        attr_value = getattr(nt, attr_name)

        # if the user specified a serializer for this attribute name, use that
        if attr_name in attr_serializers.keys():
            # serialize the value into the return dict
            json_dict[attr_name] = attr_serializers[attr_name](attr_value)
            continue
        if is_supported_container(attr_type):
            container_type, internal_type = get_collection_types(attr_type)
            # if is_optional == True, attr_value can't be None
            # if we're serialzing an non-optional, and the value is null,
            # set it to an empty container...
            # you can't pass a type_serializer (which is passed to _serialize_type)
            # to handle the internal type of a collection, if the collection is None
            # you *can* use an attr_serializer to handle the entire field, but
            # not the internal type
            if attr_value is None:
                if not is_optional:
                    warn(
                        f"No value found for non-optional type {attr_name}, defaulting to empty container"
                    )
                    json_dict[attr_name] = container_type([])
                else:
                    json_dict[attr_name] = None
                continue
            # TODO: wrap TypeError? if attr_value is iterable,
            # might not work as expected if attr_value is a string, and we iterate over chars
            json_dict[attr_name] = [
                _serialize_type(x, internal_type, False, type_serializers)
                for x in attr_value
            ]
        else:
            # single type, like:
            # a: int
            # b: Optional[str]
            # contrary to above, if attr_value here is None, we can try to use
            # any type_serializers for the attr_type that the user passed.
            # If that doesn't work, it warns the user that theres no way to
            # serialize a NoneType
            json_dict[attr_name] = _serialize_type(
                attr_value, attr_type, is_optional, type_serializers
            )
    return json_dict


def _deserialize_type(
    value: Any,
    cls: Type,
    is_optional: bool,
    type_deserializers: Dict[Type, Callable[[PrimitiveType], T]],
) -> Optional[Union[PrimitiveType, Any]]:
    """
    Gets one of the built-in deserializers or a type_deserializers from the user,
    and deserializes the loaded value to the NamedTuple representation
    """
    if cls in type_deserializers:
        return type_deserializers[cls](value)
    # is falsey value
    if value is None:
        if not is_optional:
            if type(value) != cls:
                warn(
                    f"For value {value}, expected type {cls.__name__}, found {type(value).__name__}"
                )
        return None
    elif cls == datetime:
        # serialize into epoch time
        return datetime.fromtimestamp(int(value), timezone.utc)
    elif issubclass(cls, Enum):
        return enum_getval(cls, value)
    elif cls == int:
        return int(value)
    elif cls == float:
        return float(value)
    elif cls == str:
        return str(value)
    elif cls == bool:
        if type(value) == str:
            lval = value.lower()
            if lval == "true":
                return True
            elif lval == "false":
                return False
        return bool(value)
    else:
        if is_primitive(cls):
            return value  # all other primitives are JSON compatible
        elif is_namedtuple_type(cls):
            return deserialize_namedtuple(
                value, to=cls, type_deserializers=type_deserializers
            )
    warn(f"No known way to deserialize {cls}")
    return value


def deserialize_namedtuple(
    obj: Dict[str, Any],
    to: Type[NT],
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], T]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], T]]] = None,
) -> NT:
    """
    Deserializes a Dict loaded from JSON into a NamedTuple object

    If the user provides attr_deserializers or type_deserializers, uses those
    instead of the defaults.
    """
    attr_deserializers = attr_deserializers or {}
    type_deserializers = type_deserializers or {}

    # temporary to hold values, will splat into namedtuple at the end of func
    json_dict: Dict[str, Any] = {}

    for attr_name, nt_annotation in inspect_signature_dict(to).items():
        # (<class 'int'>, False)
        attr_type, is_optional = resolve_annotation_single(nt_annotation)

        # could be None
        loaded_value: Any = obj.get(attr_name)

        # if the user specified a deserializer for this attribute name, use that
        # do attr_deserializers first, user func may have specified a way to deserialize None
        if attr_name in attr_deserializers.keys():
            json_dict[attr_name] = attr_deserializers[attr_name](loaded_value)
            continue

        # key wasnt in loaded value
        if loaded_value is None and not is_optional:
            warn(
                f"Expected key {attr_name} on non-optional field, no such key existed in loaded data"
            )

        if is_supported_container(attr_type):
            container_type, internal_type = get_collection_types(attr_type)
            # if we didnt load anything (null or key didnt exist)
            if loaded_value is None:
                if not is_optional:
                    warn(
                        f"No value loaded for non-optional type {attr_name}, defaulting to empty container"
                    )
                    json_dict[attr_name] = container_type([])
                else:
                    # else, set the optional container to none
                    # e.g. Optional[List[int]]
                    json_dict[attr_name] = None
            else:
                # if list contains nulls, _deserialize_type warns
                # its sort of up to the user how they want to use
                # - Optional[List[int]]
                # should the value be null? should it be empty list?
                # Does it somehow mean
                # - List[Optional[int]] (it shouldnt)
                # this warns in cases I think are wrong, but doesn't enforce anything
                json_dict[attr_name] = container_type(
                    [
                        _deserialize_type(
                            x, internal_type, is_optional, type_deserializers
                        )
                        for x in loaded_value
                    ]
                )
        else:
            json_dict[attr_name] = _deserialize_type(
                loaded_value, attr_type, is_optional, type_deserializers
            )
    return to(**json_dict)  # type: ignore[operator,no-any-return,call-arg]



from datetime import datetime
from typing import NamedTuple, Optional, List

from autotui import prompt_namedtuple

# describe a meeting with one or more people
class Meeting(NamedTuple):
    when: datetime
    where: Optional[str]  # asks if you want to add this
    people: List[str]  # asks if you want to add another item


m = prompt_namedtuple(Meeting)
print(m)



from pathlib import Path
from datetime import datetime
from typing import NamedTuple

from autotui.shortcuts import load_prompt_and_writeback


class Water(NamedTuple):
    at: datetime
    glasses: int
    per_glass_ml: float


def main():
    load_prompt_and_writeback(Water, "~/.local/share/water.json")
    load_prompt_and_writeback(Water, "~/.local/share/water.json")
    print(Path("~/.local/share/water.json").expanduser().read_text())


if __name__ == "__main__":
    main()



import sys
import os
import warnings
from datetime import datetime
from typing import Type, Optional, Callable, List, Union, Dict
from enum import Enum

from prompt_toolkit import prompt
from prompt_toolkit.styles import Style
from prompt_toolkit.validation import (
    Validator,
    ThreadedValidator,
    ValidationError,
    Document,
)
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.shortcuts import button_dialog, input_dialog, message_dialog

from .typehelpers import T, enum_attribute_dict

# remove pytz warning from dateparser module
warnings.filterwarnings("ignore", "The localize method is no longer necessary")

STYLE = Style.from_dict(
    {
        "dialog": "bg:#000000",
        "dialog frame.label": "bg:#ffffff #000000",
        "dialog.body": "bg:#000000 #D3D3D3",
        "dialog shadow": "bg:#000000",
        "text-area": "#000000",
    }
)


def create_repl_prompt_str(prompt_msg: str) -> str:
    """
    >>> create_repl_prompt_str("give string!")
    'give string! > '
    >>> create_repl_prompt_str("enter an int >")
    'enter an int > '
    """
    msg = prompt_msg.strip()
    if msg.endswith(">"):
        return f"{msg} "
    else:
        return f"{msg} > "


# handles the repetitive task of validating passed kwargs for prompt string for attrs
def create_prompt_string(
    for_type: Union[str, Type], for_attr: Optional[str], prompt_msg: Optional[str]
) -> str:
    # if user supplied one, use that
    pmsg = prompt_msg
    if pmsg is not None:
        return pmsg
    assert for_attr is not None, "Expected 'for_attr'; an attribute name to prompt for!"
    describe: str = for_type.__name__ if isinstance(for_type, type) else str(for_type)
    return create_repl_prompt_str(f"'{for_attr}' ({describe})")


## STRING


def prompt_str(for_attr: Optional[str] = None, prompt_msg: Optional[str] = None) -> str:

    m: str = create_prompt_string(str, for_attr, prompt_msg)
    return prompt(m)


## INT


class IntValidator(Validator):
    def validate(self, document: Document) -> None:
        text = document.text
        try:
            int(text)
        except ValueError as ve:
            raise ValidationError(message=str(ve))


def prompt_int(for_attr: Optional[str] = None, prompt_msg: Optional[str] = None) -> int:
    m: str = create_prompt_string(int, for_attr, prompt_msg)
    return int(prompt(m, validator=IntValidator()))


## FLOAT


class FloatValidator(Validator):
    def validate(self, document: Document) -> None:
        text = document.text
        try:
            float(text)
        except ValueError as ve:
            raise ValidationError(message=str(ve))


def prompt_float(
    for_attr: Optional[str] = None, prompt_msg: Optional[str] = None
) -> float:
    m: str = create_prompt_string(float, for_attr, prompt_msg)
    return float(prompt(m, validator=FloatValidator()))


## BOOL


def prompt_bool(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> bool:

    m: str = create_prompt_string(bool, for_attr, prompt_msg)
    return button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("True", True), ("False", False)],
        style=STYLE,
    ).run()


## DATETIME

DatetimeParserFunc = Callable[[str], Optional[datetime]]


class LiveDatetimeValidator(Validator):
    def __init__(
        self,
        *,
        parser_func: DatetimeParserFunc,
    ):
        super().__init__()
        self.parser_func = parser_func
        # defaults
        self.text = ""
        self.parsed: Optional[datetime] = None

    def validate(self, document: Document) -> None:
        text = document.text.strip().lower()
        self.text = text
        self.parsed = None  # reset so previous results dont stay
        if len(text) == 0:
            raise ValidationError(message="Not enough input...")
        val: Optional[datetime] = self.parser_func(text)
        if val is None:
            raise ValidationError(message=f"Couldn't parse {text} into a datetime")
        else:
            self.parsed = val

    def toolbar(self) -> str:
        result = "..."
        if len(self.text) == 0:
            return result
        if self.parsed is not None:
            result = str(self.parsed)
        else:
            result = "Couldn't parse..."
        return result


def prompt_datetime(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
) -> datetime:
    m: str = create_prompt_string(datetime, for_attr, prompt_msg)
    import dateparser  # type: ignore[import]

    # can cause lag on slower machines because of the constant
    # recomputes - put it behind a envvar-enabled feature
    if "AUTOTUI_DATETIME_LIVE" in os.environ:
        dt_validator = LiveDatetimeValidator(parser_func=dateparser.parse)
        resp = prompt(
            m,
            validator=ThreadedValidator(dt_validator),
            bottom_toolbar=dt_validator.toolbar,
        )
        dt = dateparser.parse(resp)
        assert dt is not None and isinstance(
            dt, datetime
        ), "Fatal Erorr; Could not parse response from datetime prompt into a datetime"
        return dt
    else:
        parsed_time: Optional[datetime] = None
        while parsed_time is None:
            time_str: Optional[str] = input_dialog(
                title="Describe the datetime:",
                text="For example:\n'now', '2 hours ago', 'noon', 'tomorrow at 10PM', 'may 30th at 8PM'",
                style=STYLE,
            ).run()
            if time_str is None:
                # hmm -- is this dangerous? user is prompting, so unless they've left the file
                # open this should be fine. everything in shortcuts.py is atomic-like
                # on purpose so this doesnt run into a problem there
                #
                # the alternative is to write bogus info, but perhaps that runs into
                # less problems with losing data?
                print("Cancelled, exiting...")
                sys.exit(1)
            parsed_time = dateparser.parse(time_str)
            if parsed_time is None:
                message_dialog(
                    title="Error",
                    text=f"Could not parse '{time_str}' into datetime...",
                    style=STYLE,
                ).run()
        return parsed_time


## ENUM


def _create_enum_word_targets(enum_mapping: Dict[str, Enum]) -> Dict[str, Enum]:
    # create a map of any possible description that maps back to the enumeration
    # type. When the user selects one of the descriptions, we can use this map
    # to get the corresponding Enum value
    enum_desc_map: Dict[str, Enum] = {}
    # regular key
    for k, v in enum_mapping.items():
        if k not in enum_desc_map:
            enum_desc_map[k] = v
    # casefold (lowercase) key
    for k, v in enum_mapping.items():
        kc = k.casefold()
        if kc not in enum_mapping.items():
            enum_desc_map[kc] = v
    # string representation of the value
    for k, v in enum_mapping.items():
        sv = str(v)
        if sv not in enum_desc_map:
            enum_desc_map[sv] = v
    return enum_desc_map


def prompt_enum(
    enum_cls: Type[Enum],
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> Enum:
    m: str = create_prompt_string(str, for_attr, prompt_msg)
    enum_mapping: Dict[str, Enum] = enum_attribute_dict(enum_cls)

    enum_desc_map = _create_enum_word_targets(enum_mapping)

    class EnumClosureValidator(Validator):
        def __init__(self):
            self.text = ""

        def validate(self, document: Document) -> None:
            # hmm; don't strip here since spaces might be part of the enum?
            self.text = document.text
            if self.text in enum_desc_map:
                return
            raise ValidationError(
                message=f"{self.text} is not part of the {enum_cls} enum"
            )

        def toolbar(self):
            if self.text in enum_desc_map:
                return str(enum_desc_map[self.text])
            else:
                return "..."

    validator = EnumClosureValidator()

    # prompt using a repl prompt with autocompletion/validation
    resp = prompt(
        m,
        completer=FuzzyWordCompleter(words=list(enum_desc_map)),
        validator=validator,
        bottom_toolbar=validator.toolbar,
    )

    # use what the user typed
    if resp in enum_desc_map:
        return enum_desc_map[resp]

    # else lowercase
    return enum_desc_map[resp.casefold()]


## LIST/SET repeat-prompt?


def prompt_ask_another(
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> bool:
    m = prompt_msg
    if m is None:
        assert (
            for_attr is not None
        ), "Expected 'for_attr'; an attribute name to prompt for!"
        m = f"Add another item to '{for_attr}'?"
    return button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("Yes", True), ("No", False)],
        style=STYLE,
    ).run()


## Optional


def prompt_optional(
    func: Callable[[], T],
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
    dialog_title: str = "===",
) -> Optional[T]:
    """
    A helper to ask if the user wants to enter information for an optional.
    If the user confirms, calls func (which asks the user for input)
    """
    m: Optional[str] = prompt_msg
    if m is None:
        assert (
            for_attr is not None
        ), "Expected 'for_attr'; an attribute name to prompt for!"
        m = f"'{for_attr}' is optional. Add?"
    if button_dialog(
        title=dialog_title,
        text=m,
        buttons=[("Add", True), ("Skip", False)],
        style=STYLE,
    ).run():
        return func()
    else:
        return None


##  wrap some function and display the specified thrown errors as validation errors


def prompt_wrap_error(
    func: Callable[[str], T],
    catch_errors: List[Type],
    for_attr: Optional[str] = None,
    prompt_msg: Optional[str] = None,
) -> T:
    """
    Takes the prompt string, some function which takes the string the user
    is typing as input, and possible errors to catch.

    If the function raises one of those errors, raise it as a ValidationError instead.

    This is pretty similar to prompt_toolkit.validators.Validation.from_callable
    but it allows you to specify the error message from the callable instead.
    """
    m: str = create_prompt_string(func.__name__, for_attr, prompt_msg)

    class LambdaPromptValidator(Validator):
        def validate(self, document: Document) -> None:
            text = document.text
            try:
                func(text)
            except Exception as e:
                for catchable in catch_errors:
                    if isinstance(e, catchable):
                        raise ValidationError(message=str(e))
                else:
                    # if the user didnt specify this as an error to catch
                    raise e

    return func(prompt(m, validator=LambdaPromptValidator()))



import io
from pathlib import Path
from setuptools import setup, find_packages

requirements = Path("requirements.txt").read_text().splitlines()

# Use the README.md content for the long description:
with io.open("README.md", encoding="utf-8") as fo:
    long_description = fo.read()

pkg = "autotui"
setup(
    name=pkg,
    version="0.4.0",
    url="https://github.com/seanbreckenridge/autotui",
    author="Sean Breckenridge",
    author_email="seanbrecke@gmail.com",
    description=(
        """quickly create UIs to interactively prompt, validate, and persist python objects to disk (JSON/YAML) and back using type hints"""
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(include=[pkg]),
    package_data={pkg: ["py.typed"]},
    test_suite="tests",
    install_requires=requirements,
    python_requires=">=3.8",
    extras_require={
        "testing": [
            "pytest",
            "mypy",
        ],
        "optional": ["orjson"],
    },
    keywords="data prompt namedtuple",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



import functools
from datetime import datetime
from typing import (
    Any,
    Optional,
    Union,
    List,
    Type,
    Dict,
    Callable,
)
from enum import Enum

from .typehelpers import (
    T,
    NT,
    PromptFunction,
    OptionalPromptFunction,
    PromptFunctionorValue,
    is_supported_container,
    get_collection_types,
    add_to_container,
    AllowedContainers,
    inspect_signature_dict,
    is_namedtuple_type,
    resolve_annotation_single,
)
from .warn import warn

from .exceptions import AutoTUIException


class AutoHandler:
    def __init__(
        self,
        func: Callable[[str], T],
        catch_errors: Optional[List[Type]] = None,
        prompt_msg: Optional[str] = None,
    ):
        self.func = func
        self.catch_errors: List[Type] = [] if catch_errors is None else catch_errors
        self.prompt_msg = prompt_msg


# wraps a value in a function with no arguments if needed,
# else assumes the user passed a function they wanted
# to be called, instead of prompting using prompt_toolkit
# with the attr/type validators
def _create_callable_from_user(value: PromptFunctionorValue) -> PromptFunction:
    if callable(value):
        return value  # type: ignore[return-value]
    else:
        return lambda: value  # type: ignore[return-value]


def _get_validator(
    cls: Type,
    attr_name: str,
    type_validators: Dict[Type, AutoHandler],
    type_use_values: Dict[Type, T],
) -> PromptFunction:
    """
    Gets one of the built-in validators or a type_validator from the user.
    This returns a validator for a particular type, it doesn't handle collections (List/Set)
    """
    from .prompts import (
        prompt_str,
        prompt_int,
        prompt_float,
        prompt_bool,
        prompt_datetime,
        prompt_enum,
    )

    if cls in type_use_values:
        # assuming this is a custom prompt function the user wrote, or
        # a function which returns the value to use for this
        return _create_callable_from_user(type_use_values[cls])
    elif cls in type_validators:
        return _create_callable_prompt(attr_name, type_validators[cls])
    elif cls == str:
        return lambda: prompt_str(attr_name)
    elif cls == int:
        return lambda: prompt_int(attr_name)
    elif cls == float:
        return lambda: prompt_float(attr_name)
    elif cls == bool:
        return lambda: prompt_bool(attr_name)
    elif cls == datetime:
        return lambda: prompt_datetime(attr_name)
    elif issubclass(cls, Enum):
        return lambda: prompt_enum(enum_cls=cls, for_attr=attr_name)
    # if this is another NamedTuple, call prompt_namedtuple recursively
    elif is_namedtuple_type(cls):
        return lambda: prompt_namedtuple(cls, type_validators=type_validators)
    raise AutoTUIException(f"no way to handle prompting {cls.__name__}")


# ask first would be set if is_optional was true
def _prompt_many(
    attr_name: str,
    promptfunc: PromptFunction,
    container_type: Type[AllowedContainers],
    ask_first: bool,
) -> Callable[[], AllowedContainers]:
    """
    A helper to prompt for an item zero or more times, for populating List/Set
    """
    from .prompts import prompt_ask_another

    def pm_lambda() -> AllowedContainers:
        empty_return: AllowedContainers = container_type([])
        assert isinstance(empty_return, (list, set))
        # do-while-esque
        if ask_first:
            if not prompt_ask_another(attr_name):
                return empty_return
        ret: AllowedContainers = empty_return
        continue_prompting: bool = True
        continue_ = functools.partial(
            prompt_ask_another,
            for_attr=attr_name,
            dialog_title=f"Add another item to {attr_name}?",
        )
        while continue_prompting:
            ret = add_to_container(ret, promptfunc())  # type: ignore
            # interpolate the current list into the continue? prompt
            # TODO: truncate based on terminal column width?
            continue_prompting = continue_(prompt_msg=f"Currently => {ret}")
        return ret

    return pm_lambda


def _maybe_wrap_optional(
    attr_name: str, handler: Union[AutoHandler, PromptFunction], is_optional: bool
) -> OptionalPromptFunction:
    """
    If a NamedTuple attribute is optional, wrap it
    with a dialog asking if the user wants to enter information for it
    """
    from .prompts import prompt_optional

    callf: OptionalPromptFunction = lambda: None  # dummy value
    # if user provided function/errors to catch for validation
    if isinstance(handler, AutoHandler):
        callf = _create_callable_prompt(attr_name, handler)
    else:  # is already a function
        callf = handler
    if not is_optional:
        return callf
    else:
        # if optional, wrap the typical
        # validator/callable with a yes/no prompt to add it
        return lambda: prompt_optional(func=callf, for_attr=attr_name)


def _create_callable_prompt(attr_name: str, handler: AutoHandler) -> PromptFunction:
    """
    Create a callable function with the informaton from a AutoHandler
    """
    from .prompts import prompt_wrap_error

    return lambda: prompt_wrap_error(
        func=handler.func,
        catch_errors=handler.catch_errors,
        for_attr=attr_name,
        prompt_msg=handler.prompt_msg,
    )


def _nt_dict(nt: Type, attr: str) -> Dict:
    """
    Lets the user define any of the dictionaries as a function
    which returns the dict on the class -- e.g.

    class N(NamedTuple):
        x: int
        y: str

        @staticmethod
        def type_use_values():
            return {"y": ""}  # default
    """
    d: Any = getattr(nt, attr, {})
    # if this is a function which returns the dict
    if callable(d):
        d = d()
    assert isinstance(d, dict)
    return d


# if d2 is not None, update d1 with its keys
def _update(d1: Dict, d2: Optional[Dict] = None) -> Dict:
    if d2 is not None:
        for k in d2.keys():
            d1[k] = d2[k]
    return d1


def namedtuple_prompt_funcs(
    nt: Type,
    attr_validators: Optional[Dict[str, AutoHandler]] = None,
    type_validators: Optional[Dict[Type[T], AutoHandler]] = None,
    attr_use_values: Optional[Dict[str, PromptFunctionorValue]] = None,
    type_use_values: Optional[Dict[Type[T], PromptFunctionorValue]] = None,
) -> Dict[str, PromptFunction]:
    """
    Parses the signature of a NamedTuple received from the User

    If any of the parameters cant be handled by autotui supported validators,
    checks the vaidators dict for user-supplied ones.

    Else, prints an error and fails
    """

    # if the user defined a function to return
    # the one of the dicts here, get that info
    # any kwargs passed explicitly override keys on
    # that dictionary
    # the attribute defined on it to use that for defaults
    _attr_validators = _update(_nt_dict(nt, "attr_validators"), attr_validators)
    _type_validators = _update(_nt_dict(nt, "type_validators"), type_validators)

    _attr_use_values = _update(_nt_dict(nt, "attr_use_values"), attr_use_values)
    _type_use_values = _update(_nt_dict(nt, "type_use_values"), type_use_values)

    # warn if this doesn't look like a NamedTuple
    if not is_namedtuple_type(nt):
        warn(f"{nt} doesn't look like a NamedTuple")

    # example:
    # class X(NamedTuple):
    #    a: int
    #    b: float
    #    c: str
    # >>> inspect.signature(X)
    # <Signature (a: int, b: float, c: str)>
    # the dict of attribute names -> validator (prompt) functions
    # to populate the namedtuple fields
    prompt_functions: Dict[str, PromptFunction] = {}

    # example:
    # [('a', <Parameter "a: int">), ('b', <Parameter "b: float">), ('c', <Parameter "c: str">)]
    # nt_annotation is the type
    for attr_name, nt_annotation in inspect_signature_dict(nt).items():

        if attr_name in _attr_use_values:
            prompt_functions[attr_name] = _create_callable_from_user(
                _attr_use_values[attr_name]
            )
            continue

        # (<class 'int'>, False)
        attr_type, is_optional = resolve_annotation_single(nt_annotation)

        # if the user specified a validator for this attribute name, use that
        if attr_name in _attr_validators:
            handler: AutoHandler = _attr_validators[attr_name]
            prompt_functions[attr_name] = _maybe_wrap_optional(
                attr_name, handler, is_optional
            )
            # check return type of callable to see if it matches expected type?
            continue
        promptfunc: PromptFunction
        if is_supported_container(attr_type):
            # check internal types to see if those are supported
            # optional is this context means maybe they dont want to add anything?
            # if optional present:
            #   ask before adding first item
            # else:
            #   ask after adding first item

            # e.g. if List[int], internal_type == int

            container_type, internal_type = get_collection_types(attr_type)

            # TODO: pass prompt_msg from internal type to prompt another kwarg??
            promptfunc = _get_validator(
                internal_type, attr_name, _type_validators, _type_use_values
            )
            # wrap to ask one or more times
            # wrap in container_type List/Set
            prompt_functions[attr_name] = _prompt_many(
                attr_name, promptfunc, container_type, is_optional
            )
            # TODO: recursive container support? Though if a type is getting that complicated
            # should you be using autotui in general? would be simpler to create a class and
            # supply that as a type_validator
        else:
            # single type, like:
            # a: int
            # b: Optional[str]
            promptfunc = _get_validator(
                attr_type, attr_name, _type_validators, _type_use_values
            )
            prompt_functions[attr_name] = _maybe_wrap_optional(
                attr_name, promptfunc, is_optional
            )
    # warn if no attributes are extracted
    if len(prompt_functions) == 0:
        warn("No parameters extracted from object, may not be NamedTuple?")
    return prompt_functions


def prompt_namedtuple(
    nt: Type[NT],
    *,
    attr_validators: Optional[Dict[str, AutoHandler]] = None,
    type_validators: Optional[Dict[Type[T], AutoHandler]] = None,
    attr_use_values: Optional[Dict[str, PromptFunctionorValue]] = None,
    type_use_values: Optional[Dict[Type[T], PromptFunctionorValue]] = None,
) -> NT:
    """
    Generate the list of functions using namedtuple_prompt_funcs
    and prompt the user for each of them

    attr_validators and type_validators use those functions
    to validate while prompting interactively

    attr_use_values and type_use_values let you pass default values
    to use for some attribute/type on the NamedTuple instead of prompting. the
    values for those can either be a function to call (if you wanted
    write custom code to prompt the user), or just a default value
    """

    funcs: Dict[str, PromptFunction] = namedtuple_prompt_funcs(
        nt, attr_validators, type_validators, attr_use_values, type_use_values
    )
    nt_values: Dict[str, T] = {
        attr_key: attr_func() for attr_key, attr_func in funcs.items()
    }
    return nt(**nt_values)  # type: ignore[operator, no-any-return, call-arg]



from pathlib import Path
from typing import (
    Callable,
    Type,
    Union,
    Dict,
    List,
    Any,
    Optional,
)


from . import (
    AutoHandler,
    prompt_namedtuple,
    namedtuple_sequence_dumps,
    namedtuple_sequence_load,
)
from .fileio import Format
from .typehelpers import PrimitiveType, NT, T, PromptFunctionorValue


def _normalize(_path: Union[Path, str]) -> Path:
    p: Path
    if isinstance(_path, str):
        p = Path(_path)
    else:
        p = _path
    return p.expanduser().absolute()


def _detect_format(path: Path, format: Optional[Format]) -> Optional[Format]:
    ext = path.suffix
    if ext == ".json":
        format = "json"
    elif ext in [".yml", ".yaml"]:
        format = "yaml"
    return format


# doesnt expose underlying kwargs from namedtuple_sequence_dump
# and namedtuple_sequence_load on purpose -- so that its less
# likely its mistyped. Can always use the underlying
# functions if you'd prefer to do that.


def dump_to(
    items: List[NT],
    path: Union[Path, str],
    *,
    attr_serializers: Optional[Dict[str, Callable[[T], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    format: Optional[Format] = None,
) -> None:
    """
    Takes a list of NamedTuples (or subclasses) and a path to a file.
    Serializes the items into a string, using the attr_serializers and
    type_serializers to handle custom types if specified.

    If format is unset, uses the file extension to detect the type

    If serialization succeeds, writes to the file.
    """
    p = _normalize(path)
    format = _detect_format(p, format)
    # serialize to string before opening file
    # if serialization fails, file is left alone
    nt_string: str = namedtuple_sequence_dumps(
        items,
        attr_serializers=attr_serializers,
        type_serializers=type_serializers,
        format=format,
    )
    with p.open(mode="w") as f:
        f.write(nt_string)


# args are slightly reordered here, comapared to json.load
# to be consistent with dump_to
def load_from(
    to: Type[NT],
    path: Union[Path, str],
    *,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    format: Optional[Format] = None,
    allow_empty: bool = False,
) -> List[NT]:
    """
    Takes a type to load and a path to a file containing JSON.
    Reads from the file, and deserializes the items into a list,
    using the attr_deserializers and type_deserializers to handle
    custom types if specified.

    If format is unset, uses the file extension to detect the type

    If allow_empty is True, this returns an empty list if the file
    didn't exist

    Returns the list of items read from the file.
    """
    p: Path = _normalize(path)
    format = _detect_format(p, format)
    try:
        with p.open(mode="r") as f:
            items: List[NT] = namedtuple_sequence_load(
                f,
                to,
                attr_deserializers=attr_deserializers,
                type_deserializers=type_deserializers,
                format=format,
            )
    except FileNotFoundError as fne:
        if allow_empty:
            return []
        else:
            raise fne
    return items


def load_prompt_and_writeback(
    to: Type[NT],
    path: Union[Path, str],
    *,
    create_file: bool = True,
    attr_validators: Optional[Dict[str, AutoHandler]] = None,
    type_validators: Optional[Dict[Type[T], AutoHandler]] = None,
    attr_use_values: Optional[Dict[str, PromptFunctionorValue]] = None,
    type_use_values: Optional[Dict[Type[T], PromptFunctionorValue]] = None,
    attr_serializers: Optional[Dict[str, Callable[[T], PrimitiveType]]] = None,
    type_serializers: Optional[Dict[Type, Callable[[Any], PrimitiveType]]] = None,
    attr_deserializers: Optional[Dict[str, Callable[[PrimitiveType], Any]]] = None,
    type_deserializers: Optional[Dict[Type, Callable[[PrimitiveType], Any]]] = None,
    format: Optional[Format] = None,
    prompt_function: Optional[Callable[..., NT]] = None,
) -> List[NT]:
    """
    An entry point to entire library, essentially.

    Load the NamedTuples from the JSON file specified by 'path' and 'to'

    If format is unset, uses the file extension to detect the type

    If the file doesn't exist and the create_file flag is True, ignores the
    FileNotFound error when trying to read from the file.

    After reading the NamedTuples from the file, prompts you to add a new item,
    adds that to the list it reads in, and writes back to the file

    The 'prompt_function' overrides the default 'prompt_namedtuple', if you
    want to change how to prompt/the underling functions generated by namedtuple_prompt_funcs

    accepts validators, serializers and deserializers for each step of the process.
    """
    p: Path = _normalize(path)
    format = _detect_format(p, format)
    # read from file
    items: List[NT] = []
    try:
        items = load_from(
            to,
            p,
            attr_deserializers=attr_deserializers,
            type_deserializers=type_deserializers,
            format=format,
        )
    except FileNotFoundError as fne:
        if not create_file:
            raise fne
        # warn(f"File at {p} didn't exist, using empty list")
    # prompt for the new item
    chosen_prompt_function = prompt_function or prompt_namedtuple
    new_item: NT = chosen_prompt_function(
        to,
        attr_validators=attr_validators,
        type_validators=type_validators,
        attr_use_values=attr_use_values,
        type_use_values=type_use_values,
    )
    items.append(new_item)
    # dump back to file
    dump_to(
        items,
        p,
        attr_serializers=attr_serializers,
        type_serializers=type_serializers,
        format=format,
    )
    return items



from os import environ
import warnings


def warn(message: str) -> None:
    """
    If AUTOTUI_DISABLE_WARNINGS=1 is set as an environment variable
    ignore this warning

    Otherwise, print with the normal warnings.warn call
    """
    if "AUTOTUI_DISABLE_WARNINGS" in environ:
        disabled = bool(int(environ["AUTOTUI_DISABLE_WARNINGS"]))
        if disabled:
            return
    warnings.warn(message)



from pprint import pprint
from autotui.shortcuts import load_from

# from the ./shortcuts.py file in this directory
from shortcuts import Water

if __name__ == "__main__":
    pprint(load_from(Water, "~/.local/share/water.json"))



import autotui
from typing import NamedTuple
from datetime import datetime

# something to persist to a file
class Water(NamedTuple):
    at: datetime
    glass_count: float


w = autotui.prompt_namedtuple(Water)
print(w)

s = autotui.namedtuple_sequence_dumps([w], indent=None)
print(s)

b = autotui.namedtuple_sequence_loads(s, to=Water)
print(b[0])



import sys
import warnings

# check python version
if sys.version_info[0] == 3 and sys.version_info[1] < 8:
    warnings.warn("autotui requires at least python3.8")

from .namedtuple_prompt import (
    namedtuple_prompt_funcs,
    prompt_namedtuple,
    AutoHandler,
)
from .serialize import (
    serialize_namedtuple,
    deserialize_namedtuple,
)
from .fileio import (
    namedtuple_sequence_dump,
    namedtuple_sequence_dumps,
    namedtuple_sequence_load,
    namedtuple_sequence_loads,
)
